from __future__ import annotations

from typing import Any, Iterable

from acp.helpers import (
    image_block,
    text_block,
    tool_content,
    update_agent_message,
    update_agent_thought_text,
    update_user_message,
    start_tool_call,
    update_tool_call,
)


def amp_event_to_updates(event: dict[str, Any] | Any) -> list[Any]:
    """Convert a single Amp SDK event into ACP session updates."""
    updates: list[Any] = []
    event = _ensure_dict(_normalize_recursive(event))

    if _safe_get(event, "type") == "result" and _safe_get(event, "is_error"):
        error_text = str(_safe_get(event, "error", "Unknown error"))
        updates.append(update_agent_message(text_block(f"Error: {error_text}")))
        return updates

    message = _ensure_dict(_safe_get(event, "message", {}))
    content = _safe_get(message, "content")
    if isinstance(content, str):
        updates.extend(_message_to_updates(event, [content]))
        return updates

    if isinstance(content, list):
        updates.extend(_message_to_updates(event, content))

    return updates


def _message_to_updates(event: dict[str, Any], content: Iterable[Any]) -> list[Any]:
    updates: list[Any] = []
    role = _safe_get(event, "type")
    for chunk in content:
        if isinstance(chunk, str):
            updates.append(_text_update(role, chunk))
            continue
        if not isinstance(chunk, dict):
            chunk = _ensure_dict(chunk)

        chunk_type = _safe_get(chunk, "type")
        if chunk_type == "text":
            updates.append(_text_update(role, _safe_get(chunk, "text", "")))
            continue

        if chunk_type == "thinking":
            updates.append(update_agent_thought_text(_safe_get(chunk, "thinking", "")))
            continue

        if chunk_type == "image":
            source = _ensure_dict(_safe_get(chunk, "source", {}))
            if _safe_get(source, "type") == "base64":
                updates.append(
                    update_agent_message(
                        image_block(
                            data=_safe_get(source, "data", ""),
                            mime_type=_safe_get(source, "media_type", "application/octet-stream"),
                            uri=None,
                        )
                    )
                )
            elif _safe_get(source, "type") == "url":
                updates.append(
                    update_agent_message(
                        image_block(
                            data="",
                            mime_type="",
                            uri=_safe_get(source, "url"),
                        )
                    )
                )
            continue

        if chunk_type == "tool_use":
            updates.append(
                start_tool_call(
                    tool_call_id=_safe_get(chunk, "id", ""),
                    title=_safe_get(chunk, "name") or "Tool",
                    kind="other",
                    status="pending",
                    raw_input=_safe_json(_safe_get(chunk, "input")),
                )
            )
            continue

        if chunk_type == "tool_result":
            updates.append(
                update_tool_call(
                    tool_call_id=_safe_get(chunk, "tool_use_id", ""),
                    status="failed" if _safe_get(chunk, "is_error") else "completed",
                    content=_tool_result_content(_safe_get(chunk, "content"), is_error=bool(_safe_get(chunk, "is_error"))),
                )
            )
            continue

    return updates


def _text_update(role: str | None, text: str) -> Any:
    if role == "assistant":
        return update_agent_message(text_block(text))
    return update_user_message(text_block(text))


def _tool_result_content(content: Any, *, is_error: bool) -> list[Any] | None:
    def wrap(text: str) -> str:
        if not is_error:
            return text
        return f"```\n{text}\n```"
    if isinstance(content, str) and content:
        return [tool_content(text_block(wrap(content)))]
    if isinstance(content, list) and content:
        blocks: list[Any] = []
        for item in content:
            if isinstance(item, dict) and item.get("type") == "text":
                blocks.append(tool_content(text_block(wrap(item.get("text", "")))))
        return blocks if blocks else None
    return None


def _safe_json(value: Any) -> Any:
    try:
        return jsonable(value)
    except Exception:
        return None


def jsonable(value: Any) -> Any:
    if value is None:
        return None
    if isinstance(value, (str, int, float, bool)):
        return value
    if isinstance(value, list):
        return [jsonable(v) for v in value]
    if isinstance(value, dict):
        return {str(k): jsonable(v) for k, v in value.items()}
    return str(value)


def _normalize_obj(value: Any) -> dict[str, Any]:
    if isinstance(value, dict):
        return value
    if hasattr(value, "model_dump"):
        return value.model_dump()
    if hasattr(value, "dict"):
        return value.dict()
    if hasattr(value, "__dict__"):
        return dict(value.__dict__)
    return {"content": str(value)}


def _ensure_dict(value: Any) -> dict[str, Any]:
    if isinstance(value, dict):
        return value
    if isinstance(value, str):
        return {"content": value}
    return _normalize_obj(value)


def _safe_get(obj: Any, key: str, default: Any = None) -> Any:
    try:
        if isinstance(obj, dict):
            return obj.get(key, default)
        if hasattr(obj, "get"):
            return obj.get(key, default)
        if hasattr(obj, key):
            return getattr(obj, key)
    except Exception:
        return default
    return default


def _normalize_recursive(value: Any) -> Any:
    import json
    
    if value is None or isinstance(value, (str, int, float, bool)):
        return value
    
    if isinstance(value, dict):
        return {str(k): _normalize_recursive(v) for k, v in value.items()}
    
    if isinstance(value, (list, tuple)):
        return [_normalize_recursive(v) for v in value]
    
    # Handle pydantic models
    if hasattr(value, "model_dump_json"):
        try:
            return json.loads(value.model_dump_json())
        except Exception:
            pass
    
    if hasattr(value, "model_dump"):
        try:
            return _normalize_recursive(value.model_dump(mode="json"))
        except Exception:
            pass
        try:
            return _normalize_recursive(value.model_dump())
        except Exception:
            pass
    
    if hasattr(value, "json") and callable(getattr(value, "json")):
        try:
            return json.loads(value.json())
        except Exception:
            pass
    
    if hasattr(value, "dict") and callable(getattr(value, "dict")):
        try:
            return _normalize_recursive(value.dict())
        except Exception:
            pass
    
    if hasattr(value, "__dict__"):
        return _normalize_recursive(dict(value.__dict__))
    
    return str(value)



from __future__ import annotations

from typing import Any, AsyncIterator

from acp_amp.driver.base import AmpDriver, DriverCapabilities


class PythonAmpDriver(AmpDriver):
    capabilities = DriverCapabilities(
        supports_images=True,
        supports_embedded_context=True,
        supports_mcp_http=True,
        supports_mcp_sse=True,
    )

    def __init__(self) -> None:
        try:
            from amp_sdk import execute, AmpOptions  # type: ignore
        except Exception as exc:  # pragma: no cover
            raise RuntimeError("amp-sdk is not installed") from exc
        self._execute = execute
        self._AmpOptions = AmpOptions

    async def start(self) -> None:
        return None

    async def close(self) -> None:
        return None

    async def stream_prompt(
        self,
        *,
        prompt: str,
        cwd: str | None,
        allow_all: bool,
        mcp_config: dict[str, Any] | None,
        thread_id: str | None,
        request_id: str,
    ) -> AsyncIterator[dict[str, Any]]:
        options = self._build_options(
            cwd=cwd,
            allow_all=allow_all,
            mcp_config=mcp_config,
            thread_id=thread_id,
        )
        try:
            async for message in self._execute(prompt, options=options):
                yield {"type": "event", "id": request_id, "event": _normalize_event(message)}
            yield {"type": "done", "id": request_id, "stopReason": "end_turn"}
        except Exception as exc:  # pragma: no cover
            yield {"type": "error", "id": request_id, "error": {"message": str(exc)}}

    async def cancel(self, request_id: str) -> None:
        return None

    def _build_options(
        self,
        *,
        cwd: str | None,
        allow_all: bool,
        mcp_config: dict[str, Any] | None,
        thread_id: str | None,
    ) -> Any:
        base: dict[str, Any] = {}
        if cwd:
            base["cwd"] = cwd
        if allow_all:
            base["dangerously_allow_all"] = True
            base["dangerouslyAllowAll"] = True
        if mcp_config:
            base["mcp_config"] = mcp_config
            base["mcpConfig"] = mcp_config
        if thread_id:
            base["continue"] = thread_id
            base["continue_thread"] = thread_id
            base["thread_id"] = thread_id

        candidates = [
            base,
            {k: v for k, v in base.items() if k not in ("dangerouslyAllowAll", "mcpConfig", "thread_id")},
            {k: v for k, v in base.items() if k not in ("dangerously_allow_all", "mcp_config", "continue_thread")},
            {k: v for k, v in base.items() if k in ("cwd", "dangerously_allow_all")},
            {k: v for k, v in base.items() if k in ("cwd", "dangerouslyAllowAll")},
            {},
        ]
        for attempt in candidates:
            try:
                return self._AmpOptions(**attempt)
            except TypeError:
                continue
        return self._AmpOptions()


def _normalize_event(message: Any) -> dict[str, Any]:
    import json
    
    # First try: use pydantic's JSON serialization (handles all nested models)
    if hasattr(message, "model_dump_json"):
        try:
            result = json.loads(message.model_dump_json())
            if isinstance(result, dict):
                return result
        except Exception:
            pass
    
    # Second try: use model_dump with json mode
    if hasattr(message, "model_dump"):
        try:
            result = message.model_dump(mode="json")
            if isinstance(result, dict):
                return _deep_to_dict(result)
        except Exception:
            pass
        try:
            result = message.model_dump()
            if isinstance(result, dict):
                return _deep_to_dict(result)
        except Exception:
            pass
    
    # Third try: use older pydantic .json() method
    if hasattr(message, "json") and callable(getattr(message, "json")):
        try:
            result = json.loads(message.json())
            if isinstance(result, dict):
                return result
        except Exception:
            pass
    
    # Fourth try: use .dict() method
    if hasattr(message, "dict") and callable(getattr(message, "dict")):
        try:
            result = message.dict()
            if isinstance(result, dict):
                return _deep_to_dict(result)
        except Exception:
            pass
    
    # Fifth try: manually extract attributes
    if hasattr(message, "__dict__"):
        return _deep_to_dict(dict(message.__dict__))
    
    # Last resort
    return {"type": "assistant", "message": {"content": str(message)}}


def _deep_to_dict(value: Any) -> Any:
    """Recursively convert any non-dict objects to dicts."""
    import json
    
    if value is None or isinstance(value, (str, int, float, bool)):
        return value
    
    if isinstance(value, dict):
        return {str(k): _deep_to_dict(v) for k, v in value.items()}
    
    if isinstance(value, (list, tuple)):
        return [_deep_to_dict(v) for v in value]
    
    # Handle pydantic models that might be nested
    if hasattr(value, "model_dump_json"):
        try:
            return json.loads(value.model_dump_json())
        except Exception:
            pass
    
    if hasattr(value, "model_dump"):
        try:
            return _deep_to_dict(value.model_dump(mode="json"))
        except Exception:
            pass
        try:
            return _deep_to_dict(value.model_dump())
        except Exception:
            pass
    
    if hasattr(value, "json") and callable(getattr(value, "json")):
        try:
            return json.loads(value.json())
        except Exception:
            pass
    
    if hasattr(value, "dict") and callable(getattr(value, "dict")):
        try:
            return _deep_to_dict(value.dict())
        except Exception:
            pass
    
    if hasattr(value, "__dict__"):
        return _deep_to_dict(dict(value.__dict__))
    
    # Convert anything else to string
    return str(value)



from __future__ import annotations

from pathlib import Path

_SHIM_JS = """#!/usr/bin/env node
// stdout reserved for NDJSON protocol; log to stderr.
console.log = console.error;
console.info = console.error;
console.warn = console.error;
console.debug = console.error;

import { createInterface } from \"node:readline\";
import { execute } from \"@sourcegraph/amp-sdk\";

const sessions = new Map();

function write(msg) {
  process.stdout.write(JSON.stringify(msg) + \"\\n\");
}

async function handleStart(msg) {
  const id = String(msg.id ?? \"\");
  if (!id) return;

  const controller = new AbortController();
  sessions.set(id, controller);

  const options = {
    cwd: msg.cwd || process.cwd(),
  };

  if (msg.allowAll) {
    options.dangerouslyAllowAll = true;
  }

  if (msg.mcpConfig && Object.keys(msg.mcpConfig).length > 0) {
    options.mcpConfig = msg.mcpConfig;
  }

  if (msg.threadId) {
    options.continue = msg.threadId;
  }

  let cancelled = false;

  try {
    for await (const event of execute({ prompt: msg.prompt || \"\", options, signal: controller.signal })) {
      write({ type: \"event\", id, event });
    }
  } catch (err) {
    if (controller.signal.aborted) {
      cancelled = true;
    } else {
      write({ type: \"error\", id, error: { message: String(err?.message || err) } });
      sessions.delete(id);
      return;
    }
  }

  write({ type: \"done\", id, stopReason: cancelled ? \"cancelled\" : \"end_turn\" });
  sessions.delete(id);
}

function handleCancel(msg) {
  const id = String(msg.id ?? \"\");
  const controller = sessions.get(id);
  if (controller) {
    controller.abort();
  }
}

const rl = createInterface({ input: process.stdin, crlfDelay: Infinity });
rl.on(\"line\", (line) => {
  if (!line.trim()) return;
  let msg;
  try {
    msg = JSON.parse(line);
  } catch {
    return;
  }

  if (msg.type === \"start\") {
    handleStart(msg);
    return;
  }

  if (msg.type === \"cancel\") {
    handleCancel(msg);
    return;
  }
});
"""

_SHIM_PACKAGE_JSON = """{
  \"name\": \"acp-amp-shim\",
  \"private\": true,
  \"type\": \"module\",
  \"description\": \"Node shim for the acp-amp adapter\",
  \"dependencies\": {
    \"@sourcegraph/amp-sdk\": \"^0.1.0-20251214200908-g3251f72\"
  }
}
"""


def default_shim_dir() -> Path:
    return Path.home() / ".acp-amp" / "shim"


def shim_paths(shim_dir: Path) -> tuple[Path, Path]:
    return shim_dir / "index.js", shim_dir / "package.json"


def ensure_shim_dir(shim_dir: Path) -> None:
    shim_dir.mkdir(parents=True, exist_ok=True)
    index_js, package_json = shim_paths(shim_dir)
    if not index_js.exists():
        index_js.write_text(_SHIM_JS, encoding="utf-8")
    if not package_json.exists():
        package_json.write_text(_SHIM_PACKAGE_JSON, encoding="utf-8")



from __future__ import annotations

import argparse
import asyncio
import os
from pathlib import Path

from acp.core import run_agent

from acp_amp.driver.node_sdk import NodeAmpDriver
from acp_amp.driver.python_sdk import PythonAmpDriver
from acp_amp.server import AmpAcpAgent
from acp_amp.shim import ensure_shim_dir, default_shim_dir, shim_paths


def _default_shim_path() -> Path | None:
    env_path = os.environ.get("ACP_AMP_SHIM")
    if env_path:
        return Path(env_path)

    user_shim = default_shim_dir() / "index.js"
    if user_shim.exists():
        return user_shim

    # Try repo-local layout: acp_amp/../node-shim/index.js
    here = Path(__file__).resolve()
    candidate = here.parent.parent / "node-shim" / "index.js"
    if candidate.exists():
        return candidate

    return None


def _parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser(description="ACP adapter for Amp Code")
    subparsers = parser.add_subparsers(dest="command")

    run_parser = subparsers.add_parser("run", help="Run the ACP adapter")
    run_parser.add_argument(
        "--driver",
        choices=["python", "node", "auto"],
        default=os.environ.get("ACP_AMP_DRIVER", "auto"),
        help="Amp SDK driver to use (default: auto)",
    )
    run_parser.add_argument(
        "--node",
        default=os.environ.get("ACP_AMP_NODE", "node"),
        help="Node.js executable (default: node)",
    )
    run_parser.add_argument(
        "--shim",
        default=str(_default_shim_path() or ""),
        help="Path to the Node shim (default: auto-detect or ACP_AMP_SHIM)",
    )

    setup_parser = subparsers.add_parser("setup", help="Create the local Node shim files")
    setup_parser.add_argument(
        "--path",
        default=str(default_shim_dir()),
        help="Directory to write the shim files (default: ~/.acp-amp/shim)",
    )

    return parser.parse_args()


async def _run() -> None:
    args = _parse_args()
    if args.command == "setup":
        shim_dir = Path(args.path).expanduser()
        ensure_shim_dir(shim_dir)
        index_js, package_json = shim_paths(shim_dir)
        print(f"Shim files written to {shim_dir}")
        print("Install dependencies with:")
        print(f"  cd {shim_dir} && npm install")
        print("Then run:")
        print("  acp-amp run")
        return

    if args.command in (None, "run"):
        driver_choice = args.driver if args.command == "run" else "auto"
        driver = None
        if driver_choice in ("python", "auto"):
            try:
                driver = PythonAmpDriver()
            except Exception:
                if driver_choice == "python":
                    raise
        if driver is None:
            node_cmd = args.node if args.command == "run" else os.environ.get("ACP_AMP_NODE", "node")
            shim = args.shim if args.command == "run" else str(_default_shim_path() or "")
            if not shim:
                raise SystemExit("ACP AMP: node shim not found; run `acp-amp setup` or set ACP_AMP_SHIM")
            driver = NodeAmpDriver(node_cmd=node_cmd, shim_path=Path(shim))
        agent = AmpAcpAgent(driver)
        await run_agent(agent)


def main() -> None:
    asyncio.run(_run())


if __name__ == "__main__":
    main()
